from ctypes import c_int16
from wonambi import Data
from wonambi.utils.simulate import _make_chan_name
from numpy import (append,
                   arange,
                   array,
                   insert,
                   pi,
                   save,
                   sin,
                   tan,
                   )
from bidso.utils import replace_underscore

from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus
from popeye.og import GaussianModel
from popeye.utilities import spm_hrf

from .ieeg import fake_time
from .fmri import create_bold


S_FREQ = 500
N_CHAN = 10
DUR = 1
STIM_FILE = 'prf.npy'

SCREEN_WIDTH = 25
N_PIXELS = 100
VIEWING_DISTANCE = SCREEN_WIDTH / (tan(pi / 360 * N_PIXELS) * 2)


def simulate_bold_prf(bids_dir, task_prf):
    prf_file = task_prf.get_filename(bids_dir)
    prf_file.parent.mkdir(exist_ok=True, parents=True)
    bars = generate_bars()
    stimulus = generate_stimulus(bars)
    model = generate_model(stimulus, spm_hrf)
    X = -2
    Y = 2
    SIGMA = 2
    BETA = 1
    BASELINE = 0
    dat = model.generate_prediction(X, Y, SIGMA, BETA, BASELINE)

    create_bold(prf_file, task_prf.task, 11143, dat)


def simulate_ieeg_prf(bids_dir, task_prf):
    prf_file = task_prf.get_filename(bids_dir)

    bars = generate_bars()
    stimulus = generate_stimulus(bars)
    model = generate_model(stimulus, nohrf)
    dat = generate_population_data(model, N_CHAN)

    chan = _make_chan_name(n_chan=N_CHAN)
    data = Data(data=dat, s_freq=S_FREQ, chan=chan, time=arange(dat[0].shape[0]) / S_FREQ)

    data.start_time = fake_time
    data.export(prf_file, 'bids')

    create_prf_events(replace_underscore(prf_file, 'events.tsv'), bars.shape[2])

    stimuli_dir = bids_dir / 'stimuli'
    stimuli_dir.mkdir(exist_ok=True, parents=True)
    bar_file = stimuli_dir / STIM_FILE
    save(bar_file, bars)


def generate_bars():
    thetas = arange(0, 360, 90)
    thetas = insert(thetas, 0, -1)
    thetas = append(thetas, -1)
    num_blank_steps = 30
    num_bar_steps = 30
    ecc = 12
    bar = simulate_bar_stimulus(N_PIXELS, N_PIXELS, VIEWING_DISTANCE,
                                SCREEN_WIDTH, thetas, num_bar_steps, num_blank_steps, ecc)
    return bar


def generate_stimulus(bars):
    tr_length = 1.0
    scale_factor = 1.0
    dtype = c_int16
    stimulus = VisualStimulus(bars, VIEWING_DISTANCE, SCREEN_WIDTH, scale_factor, tr_length, dtype)

    return stimulus


def nohrf(*args):
    return array([1, ])


def generate_model(stimulus, hrf_func):
    model = GaussianModel(stimulus, hrf_func)
    model.hrf_delay = 0
    model.mask_size = 6
    return model


def generate_population_data(model, n_chan):
    # generate a random pRF estimate
    X = 10
    Y = 10
    SIGMA = 2
    BETA = 1
    BASELINE = 0

    FREQ = 70
    t = arange(S_FREQ * DUR) / S_FREQ

    dat = []
    for i in range(n_chan):
        i_dat = model.generate_prediction(X, Y, SIGMA, BETA, BASELINE)
        i_dat -= i_dat.min()

        x = i_dat[:, None] * sin(2 * pi * t * FREQ)
        dat.append(x.flatten())

    return array(dat)


def create_prf_events(tsv_file, n_events):

    with tsv_file.open('w') as f:
        f.write('onset\tduration\ttrial_type\tstim_file\tstim_file_index\n')
        for i in range(n_events):
            f.write(f'{i * DUR}\t{DUR:f}\t{i + 1:d}\t{STIM_FILE}\t{i + 1:d}\n')



from .anat import simulate_anat
from .fmri import simulate_bold
from .ieeg import simulate_ieeg, simulate_electrodes
from .pipeline import simulate_all

from pathlib import Path
with (Path(__file__).parent / 'VERSION').open() as f:
    __version__ = f.read().strip()



from pathlib import Path


ROOT_DIR = Path(__file__).parent

ANAT_DIR = ROOT_DIR / 'data' / 'anat'
data_t1 = ANAT_DIR / 'T1.mgz'
data_aparc = ANAT_DIR / 'aparc.a2009s+aseg.mgz'
data_elec = ROOT_DIR / 'data' / 'electrodes.tsv'



from json import dump
from datetime import datetime
from shutil import copyfile
from numpy import ones, r_, arange
from numpy import random

from wonambi import Data

from bidso.objects import Electrodes, iEEG
from bidso.utils import replace_underscore, replace_extension, bids_mkdir

from .data import data_elec
from .fmri import create_events


S_FREQ = 256
DURATION = 192
AMPLITUDE = 1000
EFFECT_SIZE = 2
block_dur = 32
EXTRA_CHANS = ['EOG1', 'EOG2', 'ECG', 'EMG', 'other']

BV_ORIENTATION = 'MULTIPLEXED'  # 'F'
BV_DATATYPE = 'IEEE_FLOAT_32'  # float32
RESOLUTION = 1

fake_time = datetime(2017, 5, 16, 20, 45, 6)


def simulate_ieeg(root, ieeg_task, elec):
    bids_mkdir(root, ieeg_task)

    chan_names = [x['name'] for x in elec]
    ieeg_file = ieeg_task.get_filename(root)

    create_ieeg_data(ieeg_file, chan_names)
    create_ieeg_info(replace_extension(ieeg_file, '.json'))
    create_channels(replace_underscore(ieeg_file, 'channels.tsv'), elec)
    create_events(replace_underscore(ieeg_file, 'events.tsv'))

    return iEEG(ieeg_file)


def simulate_electrodes(root, elec_obj, electrodes_file=None):
    bids_mkdir(root, elec_obj)

    if electrodes_file is None:
        electrodes_file = data_elec
    output_file = elec_obj.get_filename(root)
    copyfile(electrodes_file, output_file)

    coordsystem_file = replace_underscore(output_file, 'coordsystem.json')
    COORDSYSTEM = {
        "iEEGCoordinateSystem": 'T1w',
        "iEEGCoordinateUnits": 'mm',
        "iEEGCoordinateProcessingDescription": "none",
        "IntendedFor": "/sub-bert/ses-day01/anat/sub-bert_ses-day01_acq-wholebrain_T1w.nii.gz",
        "AssociatedImageCoordinateSystem": "T1w",
        "AssociatedImageCoordinateUnits": "mm",
        }

    with coordsystem_file.open('w') as f:
        dump(COORDSYSTEM, f, indent=' ')

    return Electrodes(output_file)


def create_ieeg_data(output_file, elecs):

    n_chan = len(elecs) + len(EXTRA_CHANS)

    random.seed(100)
    t = r_[ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ), ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ), ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ)]
    dat = random.random((n_chan, S_FREQ * DURATION)) * t[None, :] * AMPLITUDE

    data = Data(dat, S_FREQ, chan=elecs + EXTRA_CHANS, time=arange(dat.shape[1]) / S_FREQ)
    data.start_time = fake_time
    data.export(output_file, 'bids')


def create_channels(output_file, elec):
    with output_file.open('w') as f:
        f.write('name\ttype\tunits\tsampling_frequency\tlow_cutoff\thigh_cutoff\tnotch\treference\tstatus\n')
        for one_elec in elec:
            f.write(f'{one_elec["name"]}\tECOG\tµV\t{S_FREQ}\tn/a\tn/a\tn/a\tn/a\tgood\n')

        for chan_name in EXTRA_CHANS:
            f.write(f'{chan_name}\tother\tµV\t{S_FREQ}\tn/a\tn/a\tn/a\tn/a\tgood\n')


def create_ieeg_info(output_file):
    """Use only required fields
    """
    dataset_info = {
        "TaskName": "block",
        "Manufacturer": "simulated",
        "PowerLineFrequency": 50,
    }

    with output_file.open('w') as f:
        dump(dataset_info, f, indent=' ')



"""BIDS infrastructure
"""
from json import dump

from .files import subject


def simulate_bids(ROOT_PATH):

    participants_tsv = ROOT_PATH / 'participants.tsv'
    with participants_tsv.open('w') as f:
        f.write('participant_id\tage\tsex\n')

        f.write(f'{subject}\t30\tF\n')

    d = {
        "Name": "Dataset with simulated data",
        "BIDSVersion": "1.1",
        "Authors": ["Giovanni Piantoni", ],
        }

    with (ROOT_PATH / 'dataset_description.json').open('w') as f:
        dump(d, f, ensure_ascii=False, indent=' ')



from json import dump
from numpy import ones, r_, random, zeros, sum, where
from skimage.transform import downscale_local_mean

from nibabel import Nifti1Image
from nibabel import load as nload

from bidso.objects import Task
from bidso.utils import replace_extension, replace_underscore, bids_mkdir

from .data import data_aparc


def simulate_bold(root, task_fmri):
    bids_mkdir(root, task_fmri)

    fmri_file = task_fmri.get_filename(root)
    SHIFT = 3
    x = r_[ones(SHIFT), ones(16) * 5, ones(16), ones(16) * 5, ones(16), ones(16) * 5, ones(16 - SHIFT)]

    create_bold(fmri_file, task_fmri.task, 11129, x)

    create_events(replace_underscore(fmri_file, 'events.tsv'))

    return Task(fmri_file)


def create_bold(bold_file, taskname, region_idx, timeseries):

    aparc = nload(str(data_aparc))

    brain = select_region(aparc, lambda x: x > 0)
    act = select_region(aparc, lambda x: x == region_idx)

    t = timeseries.shape[0]

    random.seed(100)
    idx = where(brain.get_data() == 1)
    r = random.randn(idx[0].shape[0], t)

    bold = zeros(brain.get_data().shape + (t, ))
    bold[act.get_data() == 1, :] = timeseries
    bold[idx] += r

    TR = 2.

    nifti = Nifti1Image(bold.astype('float32'), brain.affine)
    nifti.header['pixdim'][4] = TR
    nifti.header.set_xyzt_units('mm', 'sec')
    nifti.to_filename(str(bold_file))

    d = {
        'RepetitionTime': TR,
        'TaskName': taskname,
        }

    json_bold = replace_extension(bold_file, '.json')
    with json_bold.open('w') as f:
        dump(d, f, ensure_ascii=False, indent=' ')


def create_events(tsv_file):

    DUR = 32
    with tsv_file.open('w') as f:
        f.write('onset\tduration\ttrial_type\n')
        is_move = True
        for i in range(6):
            trial_type = 'move' if is_move else 'rest'
            is_move = not is_move
            f.write(f'{i * DUR}\t{DUR}\t{trial_type}\n')


def select_region(img, func, ratio=4):
    nii = img.get_data()
    selected = zeros(nii.shape)
    selected[func(nii)] = 1

    return downsample(selected, img.affine, ratio)


def downsample(dat, affine, ratio, MIN=0.25):

    dat = downscale_local_mean(dat, (ratio, ratio, ratio))
    dat[dat > MIN] = 1
    dat[dat <= MIN] = 0
    af = affine.copy()
    af[:3, :3] *= ratio
    af[:3, 3] += sum(affine[:3, :3], axis=1) * (ratio / 2 - 1 / 2)

    return Nifti1Image(dat, af)



from setuptools import setup, find_packages
from pathlib import Path

here = Path(__file__).resolve().parent
with (here / 'sanajeh' / 'VERSION').open() as f:
    VERSION = f.read().strip('\n')

with (here / 'README.rst').open(encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='sanajeh',
    version=VERSION,
    description='Simulate fMRI and iEEG data in BIDS format',
    long_description=long_description,
    url='https://github.com/gpiantoni/sanajeh',
    author="Gio Piantoni",
    author_email='sanajeh@gpiantoni.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        ],
    keywords='bids',
    packages=find_packages(exclude=('tests', )),
    install_requires=[
        'bidso',
        'nibabel',
        'wonambi',
        'popeye',
        ],
    extras_require={
        'test': [  # to run tests
            'pytest',
            'pytest-cov',
            'codecov',
            ],
        },
    package_data={
        'sanajeh': [
            'VERSION',
            'data/electrodes.tsv',
            'data/anat/T1.mgz',
            ],
    },
    )



from nibabel import load as nload
from nibabel import Nifti1Image

from bidso.files import file_Core


def simulate_anat(root, task_anat, t1):
    mri = nload(str(t1))
    x = mri.get_data()
    nifti = Nifti1Image(x, mri.affine)

    anat_path = task_anat.get_filename(root)
    anat_path.parent.mkdir(exist_ok=True, parents=True)
    nifti.to_filename(str(anat_path))

    return file_Core(anat_path)  # use the general file_Core (Task needs events.tsv)



from bidso.utils import read_tsv

from .bids import simulate_bids
from .anat import simulate_anat
from .fmri import simulate_bold
from .ieeg import simulate_ieeg, simulate_electrodes
from .simulate_prf import simulate_bold_prf, simulate_ieeg_prf
from .files import (task_anat,
                    task_fmri,
                    task_ieeg,
                    elec_ct,
                    task_fmri_prf,
                    task_ieeg_prf,
                    )
from .data import data_t1, data_elec


def simulate_all(ROOT_DIR):
    simulate_bids(ROOT_DIR)

    elec = read_tsv(data_elec)
    simulate_ieeg(ROOT_DIR, task_ieeg, elec)
    simulate_electrodes(ROOT_DIR, elec_ct)
    simulate_bold(ROOT_DIR, task_fmri)
    simulate_anat(ROOT_DIR, task_anat, data_t1)
    simulate_bold_prf(ROOT_DIR, task_fmri_prf)
    simulate_ieeg_prf(ROOT_DIR, task_ieeg_prf)



from bidso import file_Core


subject = 'bert'

task_ieeg = file_Core(
    subject=subject,
    session='day02',
    modality='ieeg',
    task='block',
    run='1',
    acquisition='clinical',
    extension='.eeg',
    )

task_fmri = file_Core(
    subject=subject,
    session='day01',
    modality='bold',
    task='block',
    run='1',
    extension='.nii.gz',
    )

task_anat = file_Core(
    subject=subject,
    session='day01',
    modality='T1w',
    extension='.nii.gz',
    )

elec_ct = file_Core(
    subject=subject,
    session='day02',
    modality='electrodes',
    acquisition='ct',
    extension='.tsv',
    )

task_fmri_prf = file_Core(
    subject=subject,
    session='day03',
    modality='bold',
    task='bairprf',
    run='1',
    extension='.nii.gz',
    )

task_ieeg_prf = file_Core(
    subject=subject,
    session='day04',
    modality='ieeg',
    task='bairprf',
    run='1',
    acquisition='clinical',
    extension='.eeg',
    )
